/*	$OpenBSD: alloc.c,v 1.6 2003/08/05 20:52:27 millert Exp $	*/
/*
 * Copyright (c) 2002 Marc Espie.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE OPENBSD PROJECT AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OPENBSD
 * PROJECT OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * area-based allocation built on malloc/free
 */

#include "sh.h"

struct link {
	struct link *prev;
	struct link *next;
};

Area *
ainit(Area *ap)
{
	ap->freelist = NULL;
	return ap;
}

void
afreeall(Area *ap)
{
	struct link *l, *l2;

	for (l = ap->freelist; l != NULL; l = l2) {
		l2 = l->next;
		free(l);
	}
	ap->freelist = NULL;
}

#define L2P(l)	( (void *)(((char *)(l)) + sizeof(struct link)) )
#define P2L(p)	( (struct link *)(((char *)(p)) - sizeof(struct link)) )

void *
alloc(size_t size, Area *ap)
{
	struct link *l;

	l = malloc(size + sizeof(struct link));
	if (l == NULL)
		internal_errorf(1, "unable to allocate memory");
	l->next = ap->freelist;
	l->prev = NULL;
	if (ap->freelist)
		ap->freelist->prev = l;
	ap->freelist = l;

	return L2P(l);
}

void *
aresize(void *ptr, size_t size, Area *ap)
{
	struct link *l, *l2, *lprev, *lnext;

	if (ptr == NULL)
		return alloc(size, ap);

	l = P2L(ptr);
	lprev = l->prev;
	lnext = l->next;

	l2 = realloc(l, size+sizeof(struct link));
	if (l2 == NULL)
		internal_errorf(1, "unable to allocate memory");
	if (lprev)
	    lprev->next = l2;
	else
	    ap->freelist = l2;
	if (lnext)
	    lnext->prev = l2;

	return L2P(l2);
}

void
afree(void *ptr, Area *ap)
{
	struct link *l;

	if (!ptr)
		return;

	l = P2L(ptr);

	if (l->prev)
		l->prev->next = l->next;
	else
		ap->freelist = l->next;
	if (l->next)
		l->next->prev = l->prev;

	free(l);
}

#ifdef TEST_ALLOC

Area a;
FILE *myout;

int
main(int argc, char **argv)
{
	char buf[1024];
	struct info {
		int size;
		void *value;
	};
	struct info info[1024 * 2];
	int size, ident;
	int lineno = 0;

	myout = stdout;
	ainit(&a);
	while (fgets(buf, sizeof(buf), stdin)) {
		lineno++;
		if (buf[0] == '\n' || buf[0] == '#')
			continue;
		if (sscanf(buf, " alloc %d = i%d", &size, &ident) == 2) {
			if (ident < 0 || ident > NELEM(info)) {
				fprintf(stderr, "bad ident (%d) on line %d\n",
					ident, lineno);
				exit(1);
			}
			info[ident].value = alloc(info[ident].size = size, &a);
			printf("%p = alloc(%d) [%d,i%d]\n", 
				info[ident].value, info[ident].size,
				lineno, ident);
			memset(info[ident].value, 1, size);
			continue;
		}
		if (sscanf(buf, " afree i%d", &ident) == 1) {
			if (ident < 0 || ident > NELEM(info)) {
				fprintf(stderr, "bad ident (%d) on line %d\n",
					ident, lineno);
				exit(1);
			}
			afree(info[ident].value, &a);
			printf("afree(%p) [%d,i%d]\n", info[ident].value,
				lineno, ident);
			continue;
		}
		if (sscanf(buf, " aresize i%d , %d", &ident, &size) == 2) {
			void *value;
			if (ident < 0 || ident > NELEM(info)) {
				fprintf(stderr, "bad ident (%d) on line %d\n",
					ident, lineno);
				exit(1);
			}
			value = info[ident].value;
			info[ident].value = aresize(value,
						    info[ident].size = size,
						    &a);
			printf("%p = aresize(%p, %d) [%d,i%d]\n", 
				info[ident].value, value, info[ident].size,
				lineno, ident);
			memset(info[ident].value, 1, size);
			continue;
		}
		if (sscanf(buf, " afreeall %d", &ident) == 1) {
			printf("afreeall() [%d]\n", lineno);
			afreeall(&a);
			memset(info, 0, sizeof(info));
			continue;
		}
		fprintf(stderr, "unrecognized line (line %d)\n",
			lineno);
		exit(1);
	}
	return 0;
}

void
aerror(Area *ap, const char *msg)
{
	printf("aerror: %s\n", msg);
	fflush(stdout);
	abort();
}
#endif /* TEST_ALLOC */
